#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Tue Dec 10 16:26:02 2019

@author: usingh

Methods and classes related to execution and logging
"""

import os
import subprocess
import time
from datetime import datetime 
from datetime import timedelta
import logging
import sys
import platform
from multiprocessing import cpu_count
from pyrpipe import pyrpipe_utils as pu
import json

class LogFormatter():
    """
    A formatter for logs
    """
    def __init__(self):
        self.start_time = time.time()
    
    def format(self, record):
      
        return "{}".format(record.getMessage())       
        #time_now=str(datetime.now())
        #elapsed_seconds = record.created - self.start_time
        #hours, remainder = divmod(elapsed_seconds, 3600)
        #minutes, seconds = divmod(remainder, 60)
        
        #message=record.getMessage()

        """OLD
        if record.name == "env":
            return "{}".format(record.getMessage())
        
        if message.startswith("#"):
            return "{}\t{}".format(record.getMessage(),time_now)
        """
        
        """
        if record.name == "cmd":
            return "{}\nStart:{}\tDuration: {:02}:{:02}:{:02}".format(record.getMessage(),time_now, int(hours), int(minutes), int(seconds) )
        elif record.name == "out":
            return "STDOUT:\n{}".format(record.getMessage())
        elif record.name == "err":      
            return "STDERR:\n{}".format(record.getMessage())
        """
        

class PyrpipeLogger():
    """
    Class to manage pyrpipe logs
    
    Attributes
    -----------
    
    env_logger: logger to log the current environment
    cmd_logger: logger to log the execution status, stdout, stderr and runtimes for each command run using execute_command()
    """
    def __init__(self):
        self.__name__="pyrpipeLogger"
        #loggers
        timestamp=str(datetime.now()).split(".")[0].replace(" ","-").replace(":","_")
        self.logger_basename=timestamp+"_pyrpipe"
        self.logs_dir=os.path.join(os.getcwd(),"pyrpipe_logs")
        if not os.path.isdir(self.logs_dir):
            os.mkdir(self.logs_dir)
        """
        self.cmd_loggerPath=os.path.join(self.logs_dir,self.logger_basename+"CMD.log")
        self.stdoutLoggerPath=os.path.join(self.logs_dir,self.logger_basename+"OUT.log")
        self.stderrLoggerPath=os.path.join(self.logs_dir,self.logger_basename+"ERR.log")
        """
        self.log_path=os.path.join(self.logs_dir,self.logger_basename+".log")
        self.envlog_path=os.path.join(self.logs_dir,self.logger_basename+"ENV.log")
        
        """
        self.cmd_logger=self.create_logger("cmd",self.cmd_loggerPath,LogFormatter(),logging.DEBUG)
        self.stdoutLogger=self.create_logger("out",self.stdoutLoggerPath,LogFormatter(),logging.DEBUG)
        self.stderrLogger=self.create_logger("err",self.stderrLoggerPath,LogFormatter(),logging.DEBUG)
        """
        formatter=LogFormatter()
        self.env_logger=self.create_logger("env",self.envlog_path,formatter,logging.DEBUG)
        self.cmd_logger=self.create_logger("cmd",self.log_path,formatter,logging.DEBUG)
        
        #self.stdoutLogger=self.create_logger("out",self.log_path,formatter,logging.DEBUG)
        #self.stderrLogger=self.create_logger("err",self.log_path,formatter,logging.DEBUG)
        
        #self.env_logger=self.create_logger("env",self.envlog_path,logging.Formatter("%(message)s"),logging.DEBUG)
        
        self.init_envlog()
        self.init_cmdlog()
 
        
    
    def create_logger(self,name,logfile,formatter,level=logging.DEBUG):
        """Creates a logger
        
        Parameters
        ----------
        
        name: str
            name of logger
        logfile: str
            file name to save logs
        formatter: formatter object
            formatter for log
        
        Returns: logger
            A logger object
        """
        #Get different loggers
        handler = logging.FileHandler(logfile)        
        handler.setFormatter(formatter)
        
        logger = logging.getLogger(name)
        logger.setLevel(level)
        logger.addHandler(handler)
        return logger
    
    def init_cmdlog(self):
        """init the cmdlog
        """
        self.cmd_logger.debug("#START LOG")
    

    def init_envlog(self):
        """init the envlog
        """
        self.env_logger.debug("#START LOG")
        
        #get current time
        #sesstime='Session information collected on {}'.format(datetime.now().strftime('%Y-%m-%d %H:%M'))
        #get os
        osInfo=platform.platform()
        #get python version
        pyver='Python ' + sys.version.replace('\n', '')
        #get cpu
        cpu=str(cpu_count())+' logical CPU cores'
        
        envDesc={'now':str(datetime.now().strftime("%y-%m-%d %H:%M:%S")),
                 'python':pyver,
                 'os':osInfo,
                 'cpu':cpu,
                 'syspath':str(sys.path),
                 'sysmodules':str(list(sys.modules.keys()))
                 }
        
        self.env_logger.debug(json.dumps(envDesc))
        """Old
        self.env_logger.debug(sesstime)
        self.env_logger.debug(pyver)
        self.env_logger.debug(osInfo)
        self.env_logger.debug(cpu)
        self.env_logger.debug("#SYS PATH")
        self.env_logger.debug("sys.path:"+str(sys.path))
        self.env_logger.debug("#SYS MODULES")
        self.env_logger.debug("sys.modules:"+str(sys.modules.keys()))
        """
        self.env_logger.debug("#PROGRAMS")
        #a list of logged programs
        self.logged_programs=[]
        

###create logger
pyrpipeLoggerObject=PyrpipeLogger()
pu.print_yellow("Logs will be saved to {}.log".format(pyrpipeLoggerObject.logger_basename))
    
"""
All functions that interact with shell are defined here. 
"""
    
##############Functions###########################

   
    
def getShellOutput(cmd,verbose=False):
    """Function to run a shell command and return returncode, stdout and stderr
    
    Parameters
    ----------
    
    cdm: list
        command to run
    verbose: bool
        to print messages
    
    :return: (returncode, stdout and stderr)
    :rtype: tuple
    """
    #not logging these commands
    log_message=" ".join(cmd)
    if verbose:
        pu.print_blue("$ "+log_message)
    try:
        result = subprocess.Popen(cmd,stdout=subprocess.PIPE,stderr=subprocess.STDOUT,)
        stdout,stderr = result.communicate()
        return(result.returncode,stdout,stderr)
    except:
        return(-1,"","Command failed to execute")

def getReturnStatus(cmd):
    """
    run a shell command and get the return status
    
    Parameters
    ----------
    
    cmd: list
        shell command in list

    :return: True is returncode is 0
    :rtype: bool
    """
    #not logging these commands
    status=getShellOutput(cmd)
    if status[0]==0:
        return True
    return False

#prints stdout in real time. optimal for huge stdout and no stderr
def execute_commandRealtime(cmd):
    """Execute shell command and print stdout in realtime.
    """
    popen = subprocess.Popen(cmd, stdout=subprocess.PIPE, universal_newlines=True)
    
    for stdout_line in iter(popen.stdout.readline, ""):
            yield stdout_line 
    popen.stdout.close()
    return_code = popen.wait()
    
    if return_code:
        raise subprocess.CalledProcessError(return_code, cmd)


def execute_command(cmd,verbose=False,quiet=False,logs=True,dryrun=False,objectid="NA",command_name=""):
    """Function to execute commands using popen. 
    All commands executed by this function can be logged and saved to pyrpipe logs.
    
    Parameters
    ----------
    
    cmd: list
        command to execute via popen in a list
    verbose: bool
        Whether to print stdout and stderr. Default: False. All stdout and stderr will be saved to logs regardless of this flag.
    quiet: bool
        Absolutely no output on screen
    logs: bool
        Log the execution 
    dryrun: bool
        If True, perform a dry run i.e. print commands to screen and log and exit
    objectid: string
        An id to be attached with the command. This is useful fo storing logs for SRA objects where object id is the SRR id.
    command_name: string
        Name of command to be save in log. If empty it is determined as the first element of the cmd list.

    :return: Return status.True is returncode is 0
    :rtype: bool
    """
    if not command_name:
        command_name=cmd[0]
    log_message=" ".join(cmd)
    
    #dryrun: print and exit
    if dryrun:
        pu.print_blue("$ "+log_message)
        #log
        #create a dict and dump as json
        logDict={'cmd':log_message,
                 'exitcode':"0",
                 'runtime':"0",
                 'starttime':"0",
                 'stdout':"dryrun",
                 'stderr':"",
                 'objectid':objectid,
                 'commandname':command_name
                }
        pyrpipeLoggerObject.cmd_logger.debug(json.dumps(logDict))
        return True
    
    if not quiet:
        pu.print_blue("$ "+log_message)
    time_start = time.time()
    starttime_str=time.strftime("%y-%m-%d %H:%M:%S", time.localtime(time.time()))
    try:
        result = subprocess.Popen(cmd,stdout=subprocess.PIPE,stderr=subprocess.STDOUT)
        stdout,stderr = result.communicate()
        #convert to string
        if stdout:
            stdout=stdout.decode("utf-8")
        else:
            stdout=""
        if stderr:
            stderr=stderr.decode("utf-8")
        else:
            stderr=""
        
        timeDiff = round(time.time() - time_start) #round to remove microsecond term
    
        if verbose:
            if stdout:
                pu.print_blue("STDOUT:\n"+stdout)
            if stderr:
                pu.print_boldred("STDERR:\n"+stderr)
        if not quiet:
            pu.print_green("Time taken:"+str(timedelta(seconds=timeDiff)))
            
                
        exitCode=result.returncode
        
        ##Add to logs        
        if logs:

            ##get the program used and log its path
            if command_name not in pyrpipeLoggerObject.logged_programs:
                ##get which thisProgram
                #if subcommands are present use parent command
                parent_command=cmd[0]
                progDesc={'name':command_name,
                          'version':getProgramVersion(parent_command).strip(),
                          'path':getProgramPath(parent_command).strip()
                          }
                pyrpipeLoggerObject.env_logger.debug(json.dumps(progDesc))
                pyrpipeLoggerObject.logged_programs.append(command_name)
            
            #create a dict and dump as json
            logDict={'cmd':log_message,
                 'exitcode':exitCode,
                 'runtime':str(timedelta(seconds=timeDiff)),
                 'starttime':str(starttime_str),
                 'stdout':stdout,
                 'stderr':stderr,
                 'objectid':objectid,
                 'commandname':command_name
                }
            pyrpipeLoggerObject.cmd_logger.debug(json.dumps(logDict))
    
        if exitCode==0:
            return True
        return False
    #handle exceptions
    except OSError as e:
        pu.print_boldred("OSError exception occured.\n"+str(e))
        #log error
        timeDiff = round(time.time() - time_start)
        logDict={'cmd':log_message,
                 'exitcode':'-1',
                 'runtime':str(timedelta(seconds=timeDiff)),
                 'starttime':str(starttime_str),
                 'stdout':"",
                 'stderr':"OSError exception occured.\n"+str(e),
                 'objectid':objectid,
                 'commandname':command_name                 
                }
        pyrpipeLoggerObject.cmd_logger.debug(json.dumps(logDict))
        return False
    except subprocess.CalledProcessError as e:
        pu.print_boldred("CalledProcessError exception occured.\n"+str(e))
        #log error
        timeDiff = round(time.time() - time_start)
        logDict={'cmd':log_message,
                 'exitcode':'-1',
                 'runtime':str(timedelta(seconds=timeDiff)),
                 'starttime':str(starttime_str),
                 'stdout':"",
                 'stderr':"CalledProcessError exception occured.\n"+str(e),
                 'objectid':objectid,
                 'commandname':command_name                 
                }
        pyrpipeLoggerObject.cmd_logger.debug(json.dumps(logDict))
        return False
    except:
        pu.print_boldred("Fatal error occured during execution.\n"+str(sys.exc_info()[0]))
        #log error
        timeDiff = round(time.time() - time_start)
        logDict={'cmd':log_message,
                 'exitcode':'-1',
                 'runtime':str(timedelta(seconds=timeDiff)),
                 'starttime':str(starttime_str),
                 'stdout':"",
                 'stderr':str("Fatal error occured during execution.\n"+str(sys.exc_info()[0])),
                 'objectid':objectid,
                 'commandname':command_name
                }
        pyrpipeLoggerObject.cmd_logger.debug(json.dumps(logDict))
        return False
    



#modified from https://www.biostars.org/p/139422/
def is_paired(sra_file):
    """Function to test wheather a .sra file is paired or single.
    
    Parameters
    ----------
    
    sra_file (string)  the path ro sra file
    
    
    :return: True is sra is paired
    :rtype: bool
    """
    if not pu.check_files_exist(sra_file):
        raise Exception("Error checking layout. {0} doesn't exist".format(sra_file));
    
    try:
        fastqdCmd=["fastq-dump","-X","1","-Z","--split-spot", sra_file]
        output = subprocess.check_output(fastqdCmd,stderr=subprocess.DEVNULL);
        numLines=output.decode("utf-8").count("\n")
        if(numLines == 4):
            return False;
        elif(numLines == 8):
            return True
        else:
            raise Exception("Unexpected output from fast-dump");
    except subprocess.CalledProcessError as e:
        raise Exception("Error running fastq-dump: {}".format(str(e)));


def getProgramPath(programName):
    """
    Get path of installed program
    Returns the path as string    
    """
    whichCmd=['which',programName]
    out = subprocess.check_output(whichCmd,universal_newlines=True)
    return out

def getProgramVersion(programName):
    """
    Get version of installed program
    return version as string
    """
    versionCommands=['--version','-version','--ver','-ver','-v','--v']
    for v in versionCommands:
        cmd=[programName,v]
        out=getShellOutput(cmd)
        if out[0]==0:
            return out[1].decode("utf-8")
    
    return ""
    
def check_dependencies(dependencies):
    """Check whether specified programs exist in the environment.
    This uses the which command to test whether a program is present.
    
    Parameters
    ----------
    
    dependencies: list
        list of programs to test

    :return: True is all dependencies are satified, False otherwise.
    :rtype: bool
    """
    errorFlag=False
    for s in dependencies:
        #print_blue("Checking "+s+"...")
        thisCmd=['which',s]
        if(getReturnStatus(thisCmd)):
            #print_green ("Found "+s)
            pass
        else:
            pu.print_boldred ("Can not find "+s)
            errorFlag=True
    if errorFlag:
        return False
    return True


#TODO: Re-implement following using native python libraries and move to utils
def deleteFileFromDisk(filePath):
    """Delete a given file from disk
    Returns true if file is deleted or doesn't exist
    """
    if pu.check_files_exist(filePath):
        rm_Cmd=['rm',filePath]
        rv= getReturnStatus(rm_Cmd)
        return rv
    #if file doesn't exist return true
    return True

def deleteMultipleFilesFromDisk(*args):
    """Delete multiple files passed as argument.
    returns true is all files a re deleted
    """
    errorFlag=False
    for filePath in args:
        status=deleteFileFromDisk(filePath)
        if not status:
            errorFlag=True
    
    return not(errorFlag)

def move_file(source,destination):
    """perform mv command to move a file from sourc to destination
    Returns True if move is successful
    """
    #print("MOV:"+source+"-->"+destination)
    mv_cmd=['mv',source,destination]
    if not getReturnStatus(mv_cmd):
        return False
    return True


#TODO: use os.scandir
def find_files(search_path,search_pattern,recursive=False,verbose=False):
    """Function to find files using find command and return as list
    Use global paths for safety
    
    Parameters
    ----------
    
    search_path: str
        path to search under
    search_pattern: str
        pattern to search e.g. "*.bam"
    recursive: bool
        search all subdirs if true

    :return: list containing the found paths
    :rtype: list
    """
    
    if recursive:
        find_cmd=['find', search_path,'-type', 'f','-name',search_pattern]   
    else:
        find_cmd=['find', search_path,'-maxdepth', '1','-type', 'f','-name',search_pattern] 
    if verbose:
        print("$ "+" ".join(find_cmd))
    st=getShellOutput(find_cmd)
    output=[]
    if st[0]==0:
        output=st[1].decode("utf-8").split("\n")
    #remove empty strings
    if '' in output:
        output.remove('')
    return output
    






#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Sat Nov 23 15:24:54 2019

@author: usingh
"""

import setuptools
import os
import sys


#exit if python 2
if sys.version_info.major != 3:
    raise EnvironmentError("""pyrpipe requires python 3.5 or higher.Please upgrade your python.""")
    
#read description
with open("README.md", "r") as fh:
    long_description = fh.read()

#read version info
cwd =os.path.abspath(os.path.dirname("__file__"))
version = {}
with open(os.path.join(cwd, "pyrpipe", "version.py")) as fp:
    exec(fp.read(), version)
version = version["__version__"]

if version is None:
    print("Error: version is missing. Exiting...", file=sys.stderr)
    sys.exit(1)



setuptools.setup(
    
    name="pyrpipe",
    version=version,
    author="Urminder Singh",
    author_email="usingh@iastate.edu",
    description="pyrpipe",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/urmi-21/pyrpipe",
    packages=setuptools.find_packages(),
    include_package_data=True,
    scripts=['scripts/pyrpipe_diagnostic.py'],
    install_requires=[line.rstrip() for line in open("requirements.txt", "rt")],
    tests_require=["pytest"],
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: Unix",
    ],
    python_requires='>=3.6',
)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Wed Dec  4 14:54:22 2019

@author: usingh
"""
from pyrpipe import pyrpipe_utils as pu
from pyrpipe import pyrpipe_engine as pe
import os
import yaml

class RNASeqTools:
    def __init__(self):
        self.category="RNASeqTools"
    

class Samtools(RNASeqTools):
    def __init__(self,**kwargs):
        self.programName="samtools"
        #check if hisat2 exists
        if not pe.check_dependencies([self.programName]):
            raise Exception("ERROR: "+ self.programName+" not found.")
        
        self.valid_args=['-b','-C','-1','-u','-h','-H','-c','-o','-U','-t','-L','-r',
                            '-R','-q','-l','-m','-f','-F','-G','-s','-M','-x','-B','-?','-S','-O','-T','-@']
        
        self.passedArgumentDict=kwargs
        
        
        
    def sam_to_bam(self,sam_file,out_dir="",out_suffix="",delete_sam=False,verbose=False,quiet=False,logs=True,objectid="NA",**kwargs):
        """Convert sam file to a bam file. 
        Output bam file will have same name as input sam.
        
        out_suffix: string
            Suffix for the output sam file
        delete_sam: bool
            delete the sam file after conversion
        verbose: bool
            Print stdout and std error
        quiet: bool
            Print nothing
        logs: bool
            Log this command to pyrpipe logs
        objectid: str
            Provide an id to attach with this command e.g. the SRR accession. This is useful for debugging, benchmarking and reports.
        kwargs: dict
            Options to pass to samtools. This will override the existing options 

        :return: Returns the path to the bam file. Returns empty string if operation failed.
        :rtype: string
        """        
        if not out_dir:            
            out_dir=pu.get_file_directory(sam_file)
        else:
            if not pu.check_paths_exist(out_dir):
                pu.mkdir(out_dir)
        
        fname=pu.get_file_basename(sam_file)
        
        #output will be out_bam
        out_bam=os.path.join(out_dir,fname+out_suffix+'.bam')
        
        newOpts={"--":(sam_file,),"-o":out_bam,"-b":""}
        mergedOpts={**kwargs,**newOpts}
        
        status=self.run_samtools("view",verbose=verbose,quiet=quiet,logs=logs,objectid=objectid,**mergedOpts)
                
        if not status:
            print("Sam to bam failed for:"+sam_file)
            return ""
        
        #check if bam file exists
        if not pu.check_files_exist(out_bam):
            return ""
        
        #delete_sam_file
        if delete_sam:
            if not pe.deleteFileFromDisk(sam_file):
                print("Error deleting sam file:"+sam_file)
                
        #return path to file
        return out_bam
        
        
        
        
    #sort bam file.output will be bam_file_sorted.bam
    def sort_bam(self,bam_file,out_dir="",out_suffix="",delete_bam=False,verbose=False,quiet=False,logs=True,objectid="NA",**kwargs):
        """Sorts an input bam file. Outpufile will end in _sorted.bam
        
        verbose: bool
            Print stdout and std error
        quiet: bool
            Print nothing
        logs: bool
            Log this command to pyrpipe logs
        objectid: str
            Provide an id to attach with this command e.g. the SRR accession. This is useful for debugging, benchmarking and reports.
        kwargs: dict
            Options to pass to samtools. This will override the existing options 

        :return: Returns path to the sorted bam file. Returns empty string if operation failed.
        :rtype: string
        
        """
        if not out_dir:
            out_dir=pu.get_file_directory(bam_file)
        else:
            if not pu.check_paths_exist(out_dir):
                pu.mkdir(out_dir)
                
        fname=pu.get_file_basename(bam_file)
        #output will be out_bam
        outSortedbam_file=os.path.join(out_dir,fname+out_suffix+'_sorted.bam')
        
        newOpts={"--":(bam_file,),"-o":outSortedbam_file}
        mergedOpts={**kwargs,**newOpts}
        
        status=self.run_samtools("sort",verbose=verbose,quiet=quiet,logs=logs,objectid=objectid,**mergedOpts)
        
        if not status:
            print("Bam sort failed for:"+bam_file)
            return ""
        
        #check if bam file exists
        if not pu.check_files_exist(outSortedbam_file):
            return ""

        if delete_bam:
            if not pe.deleteFileFromDisk(bam_file):
                print("Error deleting sam file:"+bam_file)
                
        #return path to file
        return outSortedbam_file
    
    def sam_sorted_bam(self,sam_file,out_dir="",out_suffix="",delete_sam=False,delete_bam=False,verbose=False,quiet=False,logs=True,objectid="NA",**kwargs):
        """Convert sam file to bam and sort the bam file.
        
        verbose: bool
            Print stdout and std error
        quiet: bool
            Print nothing
        logs: bool
            Log this command to pyrpipe logs
        objectid: str
            Provide an id to attach with this command e.g. the SRR accession. This is useful for debugging, benchmarking and reports.
        kwargs: dict
            Options to pass to samtools. This will override the existing options 

        :return: Returns path to the sorted bam file. Returns empty string if operation failed.
        :rtype: string
        """
        
        sam2bam_file=self.sam_to_bam(sam_file,delete_sam=delete_sam,verbose=verbose,quiet=quiet,logs=logs,objectid=objectid,**kwargs)
        
        if not sam2bam_file:
            return ""
            

        bamSorted=self.sort_bam(sam2bam_file,out_dir, out_suffix,delete_bam,verbose=verbose,quiet=quiet,logs=logs,objectid=objectid,**kwargs)
        
        if not bamSorted:
            return ""
        
        return bamSorted
    
    
    def merge_bam(self,*args,out_file="merged",out_dir="",delete_bams=False,verbose=False,quiet=False,logs=True,objectid="NA",**kwargs):
        """Merge multiple bam files into a single file
        
        Parameters
        ----------
        
        out_file: string
            Output file name to save the results. .bam will be added at the end.
        args:tuple
            Paths to bam files to combine
        out_dir: string
            Path where to save the merged bam file. Default path is the same as the first bam_file's
        verbose: bool
            Print stdout and std error
        quiet: bool
            Print nothing
        logs: bool
            Log this command to pyrpipe logs
        objectid: str
            Provide an id to attach with this command e.g. the SRR accession. This is useful for debugging, benchmarking and reports.
        kwargs: dict
            Options to pass to samtools. This will override the existing options 

        :return: Returns the path to the merged bam file.
        :rtype: string
        """
        
        if len(args) < 2:
            print("Please supply at least 2 files to merge")
            return ""
        
        if not out_dir:
            out_dir=pu.get_file_directory(args[0])
        else:
            if not pu.check_paths_exist(out_dir):
                pu.mkdir(out_dir)
        
        outMergedFile=os.path.join(out_dir,out_file+".bam")
        
        newOpts={"--":(outMergedFile,)+args}
        
        mergedOpts={**kwargs,**newOpts}
        
        status=self.run_samtools("merge",verbose=verbose,quiet=quiet,logs=logs,objectid=objectid,**mergedOpts)
        
        if not status:
            print("Bam merge failed for:"+outMergedFile)
            return ""
        
        #check if bam file exists
        if not pu.check_files_exist(outMergedFile):
            return ""
        

        if delete_bams:
            for bam_file in args:
                if not pe.deleteFileFromDisk(bam_file):
                    print("Error deleting sam file:"+bam_file)
                    
        return outMergedFile
        
        
        
    def run_samtools(self,sub_command,verbose=False,quiet=False,logs=True,objectid="NA",**kwargs):
        """A wrapper to run samtools.
        
        Parameters
        ----------
        
        sub_command: string
            sub_command to pass to samtools e.g. sort, merge etc
        arg1: dict
            arguments to pass to samtools. This will override parametrs already existing in the self.passedArgumentDict list but NOT replace them.
        verbose: bool
            Print stdout and std error
        quiet: bool
            Print nothing
        logs: bool
            Log this command to pyrpipe logs
        objectid: str
            Provide an id to attach with this command e.g. the SRR accession. This is useful for debugging, benchmarking and reports.
        kwargs: dict
            Options to pass to samtools. This will override the existing options 

        :return: Returns the status of samtools. True is passed, False if failed.
        :rtype: bool
        """
            
        #override existing arguments
        mergedArgsDict={**self.passedArgumentDict,**kwargs}
       
        samtools_cmd=['samtools',sub_command]
        #add options
        samtools_cmd.extend(pu.parse_unix_args(self.valid_args,mergedArgsDict))
                
        #start ececution
        status=pe.execute_command(samtools_cmd,verbose=verbose,quiet=quiet,logs=logs,objectid=objectid)
        if not status:
            pu.print_boldred("samtools failed")
        
        #return status
        return status
        
        
        
        
        
class Portcullis(RNASeqTools):
    def __init__(self,**kwargs):
        self.programName="portcullis"
        self.dep_list=[self.programName]
        #check if program exists
        if not pe.check_dependencies(self.dep_list):
            raise Exception("ERROR: "+ self.programName+" not found.")
        
        self.valid_args=['-t','--threads','-v','--verbose','--help','-o','-b',
                            '--bam_filter','--exon_gff','--intron_gff','--source',
                            '--force','--copy','--use_csi','--orientation','--strandedness',
                            '--separate','--extra','-r','--max_length','--canonical','--min_cov',
                            '--save_bad']
        
        self.passedArgumentDict=kwargs
        
        
    def run_portcullisFull(self,reference_fasta,bam_file,out_dir="",delete_bam=False,verbose=False,quiet=False,logs=True,objectid="NA",**kwargs):
        """
        run portculis full
        
        Parameters
        ----------
        
        reference_fasta: string
            Path to the reference fasta file
        bam_file: string
            Path to input bam file
        out_dir: string
            Path to the out put dir. current directory is not given.
        
        verbose: bool
            Print stdout and std error
        quiet: bool
            Print nothing
        logs: bool
            Log this command to pyrpipe logs
        objectid: str
            Provide an id to attach with this command e.g. the SRR accession. This is useful for debugging, benchmarking and reports.
        kwargs: dict
            Options to pass to portcullis. This will override the existing options 
        
        """
        
        if not pu.check_files_exist(reference_fasta,bam_file):
            print ("Please check input for portcullis.")
            return ""
        
        
        newOpts={"--":(reference_fasta,bam_file)}
        mergedOpts={**kwargs,**newOpts}
        #add out dir path
        if not out_dir:
            out_dir=os.path.join(os.getcwd(),"portcullis_out")
                  
        mergedOpts={**mergedOpts,**{"-o":out_dir}}
        
        status=self.run_portcullis("full",verbose=verbose,quiet=quiet,logs=logs,objectid=objectid,**mergedOpts)
        
        if not status:
            print("portcullis full failed for:"+bam_file)
            return ""
        
        #check if bam file exists
        if not pu.check_paths_exist(out_dir):
            return ""

        if delete_bam:
            if not pe.deleteFileFromDisk(bam_file):
                    print("Error deleting bam file:"+bam_file)
        
        return out_dir
    
    def run_portcullis(self,sub_command,verbose=False,quiet=False,logs=True,objectid="NA",**kwargs):
        """
        Wrapper to run portcullis.
        
        Parameters
        ----------
        
        sub_command: string
            sub_command to pass to portcullis e.g. full, prep, junc etc.
        verbose: bool
            Print stdout and std error
        quiet: bool
            Print nothing
        logs: bool
            Log this command to pyrpipe logs
        objectid: str
            Provide an id to attach with this command e.g. the SRR accession. This is useful for debugging, benchmarking and reports.
        kwargs: dict
            arguments to pass to portcullis. This will override parametrs already existing in the self.passedArgumentDict list but NOT replace them.

        :return: Returns the status of portcullis. True is passed, False if failed.
        :rtype: bool
        """
        
        
        #override existing arguments
        mergedArgsDict={**self.passedArgumentDict,**kwargs}
       
        portcullis_cmd=['portcullis',sub_command]
        #add options
        portcullis_cmd.extend(pu.parse_unix_args(self.valid_args,mergedArgsDict))
                
        #print("Executing:"+" ".join(portcullis_cmd))
        
        
        #start ececution
        status=pe.execute_command(portcullis_cmd,verbose=verbose,quiet=quiet,logs=logs,objectid=objectid)
        if not status:
            pu.print_boldred("portcullis failed")
                
        #return status
        return status
        
        
        

class Mikado(RNASeqTools):
    def __init__(self,**kwargs):
        self.programName="mikado"
        self.dep_list=[self.programName]
        #check if program exists
        if not pe.check_dependencies(self.dep_list):
            raise Exception("ERROR: "+ self.programName+" not found.")
        
        self.valid_args=[]        
        self.passedArgumentDict=kwargs
        
        
 
    def createMikadoGTFlist(self,out_file,out_dir,searchPath,searchQuery="*.gtf",strand=False):
        """Create a file to be used by mikado configure
        """
        
        files=pe.find_files(searchPath,searchQuery)
        args=files
        
        #create out dir
        if not pu.check_paths_exist(out_dir):
            pu.mkdir(out_dir)
        outFilePath=os.path.join(out_dir,out_file+".txt")
        
        
        gtfs=[]
        for l in args:
            thisName=pu.get_file_basename(l)
            if thisName:
                gtfs.append("\t".join([l,thisName,str(strand)]))
        
        f=open(outFilePath,"w")
        f.write("\n".join(gtfs))
        f.close()
        
        pu.print_green("Mikado list file written to:"+outFilePath)
        return outFilePath
                

        
    def runMikadoFull(self,listFile,genome,mode,scoring,junctions,config_out_file,blast_targets,orfs,xml,out_dir=None,verbose=False,quiet=False,logs=True,objectid="NA",**kwargs):
        """Run whole mikado pipeline
        Output will be stored to out_dir/
        """
        
        #run mikado config
        config_file=self.runMikadoConfigure(listFile,genome,mode,scoring,junctions,config_out_file,out_dir,verbose,quiet,logs,objectid,**kwargs)
        if not pu.check_files_exist(config_file):
            pu.print_boldred("Mikado configure failed")
            return False
        
        #run mikado prep
        self.runMikadoPrepare(config_file,out_dir,verbose,quiet,logs,objectid,**kwargs)
        
        #run mikado ser
        self.runMikadoSerialise(config_file,blast_targets,orfs, xml,out_dir,verbose,quiet,logs,objectid,**kwargs)
        
        #run mikado pick
        self.runMikadoPick(config_file,out_dir,verbose,quiet,logs,objectid,**kwargs)
        
        #return dir containing mikado results
        return out_dir
        
    
    
    
    def runMikadoConfigure(self,listFile,genome,mode,scoring,junctions,out_file,out_dir=None,verbose=False,quiet=False,logs=True,objectid="NA",**kwargs):
        """Wrapper to run mikado configure
        Make sure the paths in list file are global.
        
        Parameters
        ----------

        :return: Path to the created configuration file
        :rtype: string
        """
        
        #check all file exists
        if not pu.check_files_exist(listFile,genome,junctions,scoring):
            print("Please check mikado input")
            return ""
        
        #create out dir
        if out_dir==None:
            out_dir=os.getcwd()
        if not pu.check_paths_exist(out_dir):
            if not pu.mkdir(out_dir):
                raise Exception("Exception in mikado configure.")
            
        outFilePath=os.path.join(out_dir,out_file+".yaml")
        
        newOpts={"--list":listFile,"--reference":genome,"--mode":mode,"--scoring":scoring,"--junctions":junctions,"--out-dir":out_dir,"--":(outFilePath,)}
        
        #merge with kwargs
        mergedOpts={**kwargs,**newOpts}
        
        status=self.runMikado("configure",verbose=verbose,quiet=quiet,logs=logs,objectid=objectid,**mergedOpts)
        
        if not status:
            pu.print_boldred("Mikado configure failed.\nPlease make sure the paths in list file are global.")
            return ""
        
        #check if bam file exists
        if not pu.check_files_exist(outFilePath):
            return ""
        
        
        return outFilePath
        
    
    def runMikadoPrepare(self,yamlconf, out_dir=None,verbose=False,quiet=False,logs=True,objectid="NA",**kwargs):
        """Wrapper to run mikado prepare
        """
        
        #check input files exist
        if not pu.check_files_exist(yamlconf):
            print("Please check the input configuration to mikado.")
            return ""
        if not out_dir:
            out_dir=os.getcwd()

        newOpts={"--output-dir":out_dir,"--json-conf":yamlconf}
        
        #merge with kwargs
        mergedOpts={**kwargs,**newOpts}
        
        status=self.runMikado("prepare",verbose=verbose,quiet=quiet,logs=logs,objectid=objectid,**mergedOpts)
        
        if not status:
            print("Mikado prepare failed for:"+yamlconf)
            return ""
        
        #check if bam file exists
        if not pu.check_paths_exist(out_dir):
            return ""
        
        #after running mikado prep update the mikado config file to contain absolute path of mikafo_prepared.fa
        #otherwise serialize step will fail if out dirs are different
        #read the cofig.json file
        with open(yamlconf) as f:
            conf_data = yaml.safe_load(f)
        tx=conf_data['serialise']['files']['transcripts']
        conf_data['serialise']['files']['transcripts']=os.path.join(out_dir,tx)
        #update input to pick
        gtf=conf_data['pick']['files']['input']
        conf_data['pick']['files']['input']=os.path.join(out_dir,gtf)
        #write to file
        with open(yamlconf,'w') as f:
            yaml.dump(conf_data, f)
       
        
        return out_dir
        
        
        
    def runMikadoSerialise(self,yamlconf,blast_targets,orfs,xml,out_dir=None,verbose=False,quiet=False,logs=True,objectid="NA",**kwargs):
        """Wrapper to run mikado serialise
        """
        #check input files exist
        if not pu.check_files_exist(blast_targets,orfs,xml):
            print("Please check the input to mikado.")
            return ""
        if not out_dir:
            out_dir=os.getcwd()
        
        newOpts={"--json-conf":yamlconf,"--blast_targets":blast_targets,"--xml":xml,"--orfs":orfs,"--output-dir":out_dir}
        
        #merge with kwargs
        mergedOpts={**kwargs,**newOpts}
        
        
        
        status=self.runMikado("serialise",verbose=verbose,quiet=quiet,logs=logs,objectid=objectid,**mergedOpts)
        
        if not status:
            print("Mikado serialise failed for:"+yamlconf)
            return ""
        
        #check if out path exists
        if not pu.check_paths_exist(out_dir):
            return ""
        
        #after running mikado serialize update the mikado config file to contain absolute path of mikafodb
        #otherwise pick step will fail
        #read the cofig.json file
        with open(yamlconf) as f:
            conf_data = yaml.safe_load(f)
        db=conf_data['db_settings']['db']
        conf_data['db_settings']['db']=os.path.join(out_dir,pu.get_filename(db))
        
        
        #write to file
        with open(yamlconf,'w') as f:
            yaml.dump(conf_data, f)
        
        return out_dir
        
        
    def runMikadoPick(self,yamlconf,out_dir=None,verbose=False,quiet=False,logs=True,objectid="NA",**kwargs):
        """Wrapper to run mikado pick
        """
        #check input files exist
        if not pu.check_files_exist(yamlconf):
            print("Please check the input to mikado.")
            return ""
        if not out_dir:
            out_dir=os.getcwd()
        
        newOpts={"--json-conf":yamlconf,"--output-dir":out_dir}
        
        #merge with kwargs
        mergedOpts={**kwargs,**newOpts}
        
        status=self.runMikado("pick",verbose=verbose,quiet=quiet,logs=logs,objectid=objectid,**mergedOpts)
        
        if not status:
            print("Mikado pick failed for:"+yamlconf)
            return ""
        
        #check if bam file exists
        if not pu.check_paths_exist(out_dir):
            return ""
        
        return out_dir
        
        
    def runMikado(self,sub_command,verbose=False,quiet=False,logs=True,objectid="NA",**kwargs):
        """Wrapper to run mikado
        """
        
        #override existing arguments
        mergedArgsDict={**self.passedArgumentDict,**kwargs}
       
        mikado_Cmd=['mikado',sub_command]
        #add options
        mikado_Cmd.extend(pu.parse_unix_args(self.valid_args,mergedArgsDict))
                
        #print("Executing:"+" ".join(mergedArgsDict))
        
        #start ececution
        status=pe.execute_command(mikado_Cmd,verbose=verbose,quiet=quiet,logs=logs,objectid=objectid)
        if not status:
            pu.print_boldred("mikado failed")
        #return status
        return status
        
        
        
        
        
        
class Ribocode(RNASeqTools):
    def __init__(self,**kwargs):
        self.programName="RiboCode"
        self.dep_list=[self.programName]
        #check if program exists
        if not pe.check_dependencies(self.dep_list):
            raise Exception("ERROR: "+ self.programName+" not found.")
        
        self.valid_args=[]        
        self.passedArgumentDict=kwargs
        
        
    def runRibocode(self,gtf,genome,bam,l="no",outsuffix="ribocode_out",verbose=False,quiet=False,logs=True,objectid="NA"):
        """Wrapper to run ribocode in one step
        """
        
        #check input
        if not pu.check_files_exist(gtf,genome,bam):
            pu.print_boldred("Please check input files for Ribocode")
            return ""
        
        out_dir=pu.get_file_directory(gtf)
        outFile=os.path.join(out_dir,outsuffix)
        
        newOpts={"-g":gtf,"f":genome,"-r":bam,"-l":l,"-o":outFile}
        
        ribocode_Cmd=['RiboCode_onestep']
        ribocode_Cmd.extend(pu.parse_unix_args(self.valid_args,newOpts))
        
        status=pe.execute_command(ribocode_Cmd,verbose=verbose,quiet=quiet,logs=logs,objectid=objectid)
        if not status:
            pu.print_boldred("ribocode failed")
            return ""
        
        return outFile
        
        
        

        
       
        
        
        
        
        
        
        
        
        
        
        
